(* On initialise le générateur de nombres aléatoire avec une graine
   fixe. Cela permet que les caractères tirés « au hasard » soient
   toujours les même, ce qui facilite le débuggage.
   Une fois le code correct, on peut remplacer cette ligne par:
   let () = Random.self_init ()
   pour obtenir une grille différente à chaque exécution.
*)

let () = Random.init 13

(* La fonction [make_grid n]
   Doit créer un tableau de taille n
   de chaînes de caractères aléatoires de taille n,
   où chaque caractère est compris entre 'A' et 'Z'
*)

let make_grid n =
  Array.init n (fun _ -> String.init n (fun _ -> Char.chr (65 + Random.int 26)))

(* La fonction [print_grid grid]*)
let print_grid g =
  Array.iter (fun s -> Printf.printf "%s\n" s) g

(* La fonction load_dico charge le fichier f
   et renvoie un trie.
   Ce dernier étant utilisé comme un ensemble, on peut
   stocker une valeur bidon pour chaque clé, par
   exemple la valeur ()
   Attention, les fonctions sur les trie sont
   dans le module Trie, il faut donc utiliser ce nom de module,
   par exemple [Trie.add]
*)
let load_dico f =
  let c_in = open_in f in
  let words = String.split_on_char '\n' (In_channel.input_all c_in) in
  close_in c_in;
  List.fold_left (fun acc w -> Trie.add w () acc) Trie.empty words

(* La fonction [print_word l] prend une liste de caractères inversée,
   la renverse et crée une chaine avec.
   CF la fonction rev_implode du TP précédent.
*)
let print_word l =
  let w = Trie.rev_implode l in
  Printf.printf "%s\n" w

module Pos =
struct
  type t = (int * int)
  let compare p1 p2 = compare p1 p2
end

module PosSet = Set.Make(Pos)


(* La fonction next_positions renvoie la liste des couples
   de position voisines valide pour la grille données et qui
   n'ont pas déjà été visitées.

   Pour une case "centrale" il s'agit des 8 cases autour de la
   case donnée. Mais pour une case dans un coin ou sur les bords,
   il y a moins de voisins.
*)

let next_positions grid row col visited =
  let pos = [ (-1, -1); (-1, 0); (-1, 1);
              (0, -1);           (0, 1);
              (1, -1); (1, 0);   (1, 1);
            ]
  in
  List.fold_left (fun acc (i, j) ->
      let rr = row + i in
      let cc = col + j in
      if rr >= 0 && cc >= 0
         && rr < Array.length grid
         && cc < String.length grid.(rr)
         && not (PosSet.mem (rr, cc) visited)
      then
        (rr, cc) :: acc
      else acc
    ) [] pos


(*
  grid : la grille
  row, col : la case courante
  l : une liste de couples (char, trie)
  word: la liste inversée des lettres, la tête
  de la liste est grid.(row).[col]

  *)

(* La fonction récursive :

*)

let rec solve_list  grid row col l word visited =
  let c = grid.(row).[col] in
  match List.assoc_opt c l with
    None -> ()   (* pas de caractère trouvé *)
  | Some t -> (* un caractère trouvé, on l'ajoute à la liste et on visite le trie associé *)
    solve_node grid row col t (c::word) visited

and solve_node grid row col t word visited =
  (* on est arrivé sur le la case row col, son caractère
     est en tete de word
  *)
  let Trie.Node(o, l) = t in
  let () = match o with
      None -> () (* on ne fait rien *)
    | Some _ -> print_word word  (* on est sur un nœud terminal *)
  in
  let visited' = PosSet.add (row, col) visited in (* on ajoute cette case à l'ensemle des visités *)
  let next = next_positions grid row col visited' in (* on demande la liste des cases voisines *)
  List.iter (fun (row', col') ->
      solve_list grid row' col' l word visited'  (* pour chaque case voisine, on cherche dans la liste
                                                    du trie s'il y a un sous-trie avec le caractère de la case*)
    ) next


(* *)
let main () =
  let dico = load_dico "french_simple.txt" in
  let n = 5 in
  let grid = make_grid n in
  let () = print_grid grid in
  let Trie.Node (_, l) = dico in (* on extrait la liste des premières lettres*)
  for i = 0 to n - 1 do
    for j = 0 to n - 1 do
      solve_list grid i j l [] PosSet.empty
    done
  done


(* attention, le main est appelé tout en bas du fichier *)

module Partie2 =
struct
  (* ce qui est modifié dans la partie 2 *)
  let rec solve_list f grid row col l word visited =
    let c = grid.(row).[col] in
    match List.assoc_opt c l with
      None -> ()   (* pas de caractère trouvé *)
    | Some t -> (* un caractère trouvé, on l'ajoute à la liste et on visite le trie associé *)
      solve_node f grid row col t (c::word) visited

  and solve_node f grid row col t word visited =
    (* on est arrivé sur le la case row col, son caractère
       est en tete de word
    *)
    let Trie.Node(o, l) = t in
    let () = match o with
        None -> () (* on ne fait rien *)
      | Some _ -> f word  (* on est sur un nœud terminal *)
    in
    let visited' = PosSet.add (row, col) visited in (* on ajoute cette case à l'ensemle des visités *)
    let next = next_positions grid row col visited' in (* on demande la liste des cases voisines *)
    List.iter (fun (row', col') ->
        solve_list f grid row' col' l word visited'  (* pour chaque case voisine, on cherche dans la liste
                                                        du trie s'il y a un sous-trie avec le caractère de la case*)
      ) next



  let main_2_1 () =
    let dico = load_dico "french_simple.txt" in
    let n = 5 in
    let grid = make_grid n in
    let () = print_grid grid in
    let Trie.Node (_, l) = dico in (* on extrait la liste des premières lettres*)
    for i = 0 to n - 1 do
      for j = 0 to n - 1 do
        solve_list print_word grid i j l [] PosSet.empty
      done
    done


  let main_2_2 () =
    let dico = load_dico "french_simple.txt" in
    let n = 20 in (* question 2.3 *)
    let grid = make_grid n in
    let () = print_grid grid in

    let num_words = ref 0 in (* on crée une référence *)
    let count _ = num_words := !num_words +1 in
    (* la fonction count ignore son argument et incrémente la référence *)
    let Trie.Node (_, l) = dico in (* on extrait la liste des premières lettres*)
    for i = 0 to n - 1 do
      for j = 0 to n - 1 do
        solve_list count grid i j l [] PosSet.empty  (* count sera appelée pour chaque mot *)
      done
    done;
    Printf.printf "%d\n" !num_words (* on affiche le résultat *)

  let main_2_4 () =
    let dico = load_dico "french_simple.txt" in
    let n = 1000 in (* question 2.3 *)
    let grid = make_grid n in
    let () = print_grid grid in

    let seen_words = Hashtbl.create 16 in (* on crée une table de hachage des mots déja vu,
                                             on pourrait utiliser une ref vers un trie ou un set...*)

    let count l =
      match l with
      | [] | [ _ ]| [ _; _ ] -> () (* mot de 0, 1, 2 lettre, on ne fait rien*)
      | _ ->
        let word = Trie.rev_implode l in
        if not (Hashtbl.mem seen_words word) then begin(* on a jamais vu ce mot *)
          Hashtbl.add seen_words word () (* encore une fois on ajoute une valeur bidon *)
        end
    in
    (* la fonction count ignore son argument et incrémente la référence *)
    let Trie.Node (_, l) = dico in (* on extrait la liste des premières lettres*)
    for i = 0 to n - 1 do
      for j = 0 to n - 1 do
        solve_list count grid i j l [] PosSet.empty  (* count sera appelée pour chaque mot *)
      done
    done;
    Printf.printf "%d\n" (Hashtbl.length seen_words) (* on affiche le résultat *)

end



module Partie2_5 =
struct
  (* ce qui est modifié dans la partie 3 *)
  let rec solve_list f grid row col l word visited =
    let c = grid.(row).[col] in
    match List.assoc_opt c l with
      None -> ()
    | Some t ->
      solve_node f grid row col t ((c,(row,col))::word) visited (* on ajoute la position en plus *)

  and solve_node f grid row col t word visited =
    let Trie.Node(o, l) = t in
    let () = match o with
        None -> () (* on ne fait rien *)
      | Some _ -> f word  (* on est sur un nœud terminal *)
    in
    let visited' = PosSet.add (row, col) visited in
    let next = next_positions grid row col visited' in
    List.iter (fun (row', col') ->
        solve_list f grid row' col' l word visited'
      ) next

  let green = "\x1b[102m"
  let cyan = "\x1b[106m"

  let reset = "\x1b[0m"
  let clear_screen = "\x1b[2J\x1b[H"

  (* la première case est mise en bleu, les autres en vert *)

  let print_full (r0,c0) grid seen_words l =
    let word = Trie.rev_implode (List.map fst l) in
    if String.length word < 3 || Hashtbl.mem seen_words word then ()
    else begin
      Hashtbl.add seen_words word ();
      let pos = List.fold_left (fun acc (_, p)-> PosSet.add p acc) PosSet.empty l in
      (* on crée l'ensemble des positions des lettres du mot *)
      (* On affiche la grille caractère par caractère: *)
      Printf.printf "%s%!" clear_screen; (* on efface l'écran, %! permet de flusher la sortie *)
      for i = 0 to Array.length grid - 1 do
        for j = 0 to String.length grid.(i) - 1 do
          if i = r0 && j = c0 then
            Printf.printf "%s%c%s" cyan grid.(i).[j] reset
          else
          if PosSet.mem (i, j) pos then
            Printf.printf "%s%c%s" green grid.(i).[j] reset
          else
            Printf.printf "%c" grid.(i).[j];
        done;
        Printf.printf "\n";
      done;
      Printf.printf "\n\n%s\n%!" word;(* on affiche le mot sous la grille *)
      Unix.sleepf 0.5 (* on fait une pause entre chaque mot pour voir quelque chose *)
    end

  let main_2_5 () =
    let dico = load_dico "french_simple.txt" in
    let n = 10 in
    let grid = make_grid n in
    let () = print_grid grid in
    let Trie.Node (_, l) = dico in (* on extrait la liste des premières lettres*)
    for i = 0 to n - 1 do
      for j = 0 to n - 1 do
        solve_list (print_full (i,j) grid (Hashtbl.create 16)) grid i j l [] PosSet.empty
      done
    done

end

(*
let () = main () (* partie 1*)
let () = Partie2.main_2_1 ()
let () = Partie2.main_2_2 ()
*)
let () = Partie2_5.main_2_5 ()


(*
let () = Partie2.main_2_4 ()
*)